/*
 * Copyright (c) 1996, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "fxrepton.h"
#include "gameengine.h"

GameEngine::GameEngine()
{
	paused = false;
	sounds = true;
	width = height = 0;
	data = 0;
	status = Status_Idle;
	showError = 0;
	playSound = 0;
	showAnimation = 0;
}

GameEngine::~GameEngine()
{
	clear();
}

void GameEngine::setErrorFunc(void (*errfnc)(const char *fmt, ...))
{
	showError = errfnc;
}

void GameEngine::setSoundFunc(void (*sndfnc)(Sound snd))
{
	playSound = sndfnc;
}

void GameEngine::setAnimationFunc(void (*animfnc)(Animation anim))
{
	showAnimation = animfnc;
}

void GameEngine::clear()
{
	deleteTransporters();
	deleteMonsters();
	delete [] data;
	data = 0;
	status = Status_Idle;
	paused = false;
}

void GameEngine::loadData(const char *filename, bool resetscore)
{
	clear();

	FILE *fin = fopen(filename, "r");
	if (!fin) {
		if (showError) showError("Cannot open %s", filename);
	  	return;
  	}

	/* TODO (#1#): Lame error checking */
 	char buf[1024];
	int num_transporters, readonly;
	fgets(buf, 1024, fin);
	sscanf(buf, "%d %d %d %d %d", &width, &height, &num_transporters, &game_time, &readonly);
	fgets(buf, 1024, fin);
 	password = FXString(buf).trim();
	game_time *= 1000; // Milliseconds
	capsule_time = game_time;

	while (num_transporters--) {
		int ox, oy, dx, dy;
		fgets(buf, 1024, fin);
		sscanf(buf, "%d %d %d %d", &ox, &oy, &dx, &dy);
		addTransporter(ox, oy, dx, dy);
	}

	data = new SpriteObj[width * height];

	num_diamonds = 0;
 	if (resetscore)
  		score = 0;
	have_key = false;
	sprite = Obj_Repton;

	bool got_position = false;
	for (int y = 0; y < height; ++y) {
		fgets(buf, 1024, fin);
		for (int x = 0; x < width; ++x) {
			char c = buf[x];
			ref(x, y) = (SpriteObj)(c - 'A');
			switch (ref(x, y)) {
			case Obj_Spirit:
				ref(x, y) = Obj_Empty;
				addMonster(x, y, Obj_Spirit);
				break;
			case Obj_Strongbox:
			case Obj_Cage:
			case Obj_Diamond:
				++num_diamonds;
				break;
			case Obj_Fungus:
				addMonster(x, y, Obj_Fungus);
				break;
			case Obj_Repton:
				if (got_position) {
    					if (showError) showError("Duplicate Repton position");
					fclose(fin);
				  	clear();
				  	return;
  				}
				got_position = true;
				xpos = x;
				ypos = y;
				break;
			default:;	// Avoid warning
			}
		}
	}

	fgets(buf, 1024, fin);
 	comments = FXString(buf).trim();
	fclose(fin);

	if (!got_position) {
		if (showError) showError("No Repton position found");
	  	clear();
	  	return;
  	}

	prepareMonsters();
}

bool GameEngine::isDataLoaded()
{
	return data != 0;
}

void GameEngine::addTransporter(int ox, int oy, int dx, int dy)
{
	Transporter *t = new Transporter(ox, oy, dx, dy);
	transporters.push_back(t);
}

Transporter *GameEngine::findTransporter(int x, int y)
{
	typedef vector<Transporter *>::const_iterator I;
	for (I i = transporters.begin(); i != transporters.end(); ++i) {
		Transporter *t = *i;
		if (t->orig_x == x && t->orig_y == y)
			return t;
	}
//	if (showError) showError("Invalid transporter");
	return 0;
}

void GameEngine::deleteTransporters()
{
	typedef vector<Transporter *>::iterator I;
	for (I i = transporters.begin(); i != transporters.end(); ++i)
		delete *i;
	transporters.clear();
}

void GameEngine::addMonster(int x, int y, SpriteObj type)
{
	Monster *m = new Monster(x, y, type);
	monsters.push_back(m);
}

void GameEngine::prepareMonsters()
{
	typedef vector<Monster *>::iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i) {
		Monster *m = *i;
		int x = m->xpos, y = m->ypos;
		if (m->type == Obj_Spirit) {
			if (!isempty(x-1, y) && isempty(x, y-1))
				m->orientation = North;
			else if (!isempty(x, y-1) && isempty(x+1, y))
				m->orientation = East;
			else if (!isempty(x+1, y) && isempty(x, y+1))
				m->orientation = South;
			else if (!isempty(x, y+1) && isempty(x-1, y))
				m->orientation = West;
		}
	}
}

Monster *GameEngine::findMonster(int x, int y)
{
	typedef vector<Monster *>::const_iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i) {
		Monster *m = *i;
		if (!m->dead && m->xpos == x && m->ypos == y)
			return m;
	}
	return 0;
}

void GameEngine::deleteMonsters()
{
	typedef vector<Monster *>::iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i)
		delete *i;
	monsters.clear();
}

// Try falling down.
bool GameEngine::tryFallDown(int x, int y)
{
	SpriteObj obj = xy(x, y);

	if (xy(x, y+1) != Obj_Empty)
		return false;

	ref(x, y) = Obj_Empty;
	Monster *mp;
	if ((mp = findMonster(x, y+1)) != 0 && mp->type == Obj_Monster) {
		if (playSound) playSound(Sound_Killed_Monster);
		score += 20;
		killMonster(mp);
	}
	if (obj == Obj_Egg)
		ref(x, y+1) = Obj_Falling_Egg;
	else
		ref(x, y+1) = obj;
	if (xy(x, y+2) == Obj_Wall && (obj == Obj_Egg || obj == Obj_Falling_Egg)) {
		if (playSound) playSound(Sound_Broken_Egg);
		ref(x, y+1) = Obj_Broken_Egg;
	}
	if (xy(x, y+2) == Obj_Repton) {
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
	}

	return true;
}

// Check if the specified object can fall to the left or to the right.
bool GameEngine::tryFallHorizontal(int x, int y, int xd)
{
	SpriteObj obj = xy(x, y);

	if (xy(x+xd, y+1) == Obj_Empty && xy(x+xd, y) == Obj_Empty) {
		ref(x, y) = Obj_Empty;
		if (obj == Obj_Egg)
			ref(x+xd, y) = Obj_Falling_Egg;
		else
			ref(x+xd, y) = obj;
		return true;
	}

	return false;
}

void GameEngine::tryFallLeftOrRight(int x, int y)
{
	SpriteObj obj = xy(x, y);

	switch(xy(x, y+1)) {
	case Obj_Rock:
	case Obj_Skull:
	case Obj_Diamond:
	case Obj_Egg:
	case Obj_Key:
	case Obj_Bomb:
	case Obj_Broken_Egg:
	case Obj_Broken_Egg1:
	case Obj_Broken_Egg2:
	case Obj_Broken_Egg3:
		// Curved objects.  Rocks and eggs can fall onto these objects.
		// If can't fall to the left, fall to the right.
		if (!tryFallHorizontal(x, y, -1))
			if (!tryFallHorizontal(x, y, 1) && obj == Obj_Falling_Egg) {
				if (playSound) playSound(Sound_Broken_Egg);
				ref(x, y) = Obj_Broken_Egg;
			}
		break;
	case Obj_Wall_North_West:
	case Obj_Filled_Wall_North_West:
		if (!tryFallHorizontal(x, y, -1) && obj == Obj_Falling_Egg) {
			if (playSound) playSound(Sound_Broken_Egg);
			ref(x, y) = Obj_Broken_Egg;
		}
		break;		
	case Obj_Wall_North_East:
	case Obj_Filled_Wall_North_East:
		if (!tryFallHorizontal(x, y, 1) && obj == Obj_Falling_Egg) {
			if (playSound) playSound(Sound_Broken_Egg);
			ref(x, y) = Obj_Broken_Egg;
		}
		break;
	default:
		if (obj == Obj_Falling_Egg) {
			if (playSound) playSound(Sound_Broken_Egg);
			ref(x, y) = Obj_Broken_Egg;
		}
	}
}

// Check if the specified object can fall.
void GameEngine::tryFall(int x, int y)
{
	if (!tryFallDown(x, y))
		tryFallLeftOrRight(x, y);
}

// Check if the Spirit can go into a direction.
bool GameEngine::spiritTryDirection(Monster *m, Orientation dir, int xd, int yd)
{
	switch (xy(m->xpos + xd, m->ypos + yd)) {
	case Obj_Empty:
	case Obj_Ground1:
	case Obj_Ground2:
		m->xpos += xd;
		m->ypos += yd;
		m->orientation = dir;
		return true;
	case Obj_Cage:
		cageSpirit(m, m->xpos + xd, m->ypos + yd);
		return true;
	case Obj_Repton:
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
		return true;
	default:;	// Avoid warning
	}

	return false;
}

// Check and move the Spirit.
void GameEngine::updateSpirit(Monster *m)
{
	int mx = m->xpos, my = m->ypos;

	switch (m->orientation) {
	case North:
		if (spiritTryDirection(m, West, -1, 0))
			;
		else if (!spiritTryDirection(m, North, 0, -1)) {
			m->orientation = East;
			updateSpirit(m);
			return;
		}
		break;
	case South:
		if (spiritTryDirection(m, East, 1, 0))
			;
		else if (!spiritTryDirection(m, South, 0, 1)) {
			m->orientation = West;
			updateSpirit(m);
			return;
		}
		break;
	case East:
		if (spiritTryDirection(m, North, 0, -1))
			;
		else if (!spiritTryDirection(m, East, 1, 0)) {
			m->orientation = South;
			updateSpirit(m);
			return;
		}
		break;
	case West:
		if (spiritTryDirection(m, South, 0, 1))
			;
		else if (!spiritTryDirection(m, West, -1, 0)) {
			m->orientation = North;
			updateSpirit(m);
			return;
		}
		break;
	}

	if (m->xpos != mx || m->ypos != my)
		m->sprite = (m->sprite == Obj_Spirit) ? Obj_Spirit2 : Obj_Spirit;
}

// Check if the monster can go into a direction and the distance between
// it and repton.
double GameEngine::monsterTryDirection(Monster *m, int xd, int yd)
{
	switch (xy(m->xpos + xd, m->ypos + yd)) {
	case Obj_Empty:
	case Obj_Ground1:
	case Obj_Ground2:
		return sqrt(pow(m->xpos + xd - xpos, 2) +
			    pow(m->ypos + yd - ypos, 2));
	case Obj_Repton:
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
	default:;
	}

	return 666;	// A rathen large number
}

// Check the repton position and move the monster according with it.
void GameEngine::updateMonster(Monster *m)
{
	int xd = 0, yd = 0;
	double i, len;

	len = sqrt(pow(m->xpos - xpos, 2) + pow(m->ypos - ypos, 2));

	if ((i = monsterTryDirection(m, 0, -1)) < len) {
		xd = 0; yd = -1; len = i;
	}
	if ((i = monsterTryDirection(m, 0, 1)) < len) {
		xd = 0; yd = 1; len = i;
	}
	if ((i = monsterTryDirection(m, 1, 0)) < len) {
		xd = 1; yd = 0; len = i;
	}
	if (monsterTryDirection(m, -1, 0) < len) {
		xd = -1; yd = 0;
	}
	m->xpos += xd;
	m->ypos += yd;

	m->sprite = (m->sprite == Obj_Monster) ? Obj_Monster2 : Obj_Monster;
}

// Check the broken eggs.
void GameEngine::updateEggs()
{
	for (int y = height - 1; y >= 0; y--)
		for (int x = width - 1; x >= 0; x--)
			switch (xy(x, y)) {
			case Obj_Broken_Egg:
				ref(x, y) = Obj_Broken_Egg1;
				break;
			case Obj_Broken_Egg1:
				ref(x, y) = Obj_Broken_Egg2;
				break;
			case Obj_Broken_Egg2:
				ref(x, y) = Obj_Broken_Egg3;
				break;
			case Obj_Broken_Egg3:
				ref(x, y) = Obj_Empty;
				addMonster(x, y, Obj_Monster);
				break;
			default:;	// Avoid warning
			}
}

// Check the moveable objects (rocks and eggs).
void GameEngine::updateObjects()
{
	// Note the ``height - 2'' value.
	for (int y = height - 2; y >= 0; y--)
		for (int x = width - 1; x >= 0; x--)
			switch(xy(x, y)) {
			case Obj_Rock:
			case Obj_Egg:
			case Obj_Falling_Egg:
			case Obj_Broken_Egg:
			case Obj_Broken_Egg1:
			case Obj_Broken_Egg2:
			case Obj_Broken_Egg3:
				// Check if the object can fall.
				tryFall(x, y);
				break;
			default:;	// Avoid warning
			}
}

// Try to reproduce the plant.
bool GameEngine::tryFungus(int x, int y, int dx, int dy)
{
	if (xy(x+dx, y+dy) == Obj_Empty || xy(x+dx, y+dy) == Obj_Repton) {
		if (xy(x+dx, y+dy) == Obj_Repton) {
			if (playSound) playSound(Sound_Died);
			if (showAnimation) showAnimation(Animation_Died);
			status = Status_Died;
		}
		ref(x+dx, y+dy) = Obj_Fungus;
		addMonster(x+dx, y+dy, Obj_Fungus);
		return true;
	}

	return false;
}

// Check the funguses and reproduce.
bool GameEngine::updateFungus(Monster *m)
{
	if (rand() % 10 == 0) {
		if (tryFungus(m->xpos, m->ypos, 1, 0))
			return true;
		else if (tryFungus(m->xpos, m->ypos, -1, 0))
			return true;
		else if (tryFungus(m->xpos, m->ypos, 0, 1))
			return true;
		else if (tryFungus(m->xpos, m->ypos, 0, -1))
			return true;
	}
	return false;
}

// Reproduce the funguses
void GameEngine::updateFunguses()
{
	typedef vector<Monster *>::iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i) {
		Monster *m = *i;
		if (m->type == Obj_Fungus)
			    if (updateFungus(m))
				    break;
	}
}

// Check the monsters status and move them.
void GameEngine::updateMonsters()
{
	typedef vector<Monster *>::iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i) {
		Monster *mp = *i;
		if (mp->dead)
			continue;
		switch (mp->type) {
		case Obj_Spirit:
			updateSpirit(mp);
			break;
		case Obj_Monster:
			updateMonster(mp);
			break;
		default:;	// Avoid warning
		}
	}
}

// Move Repton vertically.
void GameEngine::moveVerticalDirect(int yd)
{
	// Move repton up or down according with the yd value.
	ref(xpos, ypos) = Obj_Empty;
	ref(xpos, ypos + yd) = Obj_Repton;
	ypos += yd;

	// Change the repton sprite orientation.
	switch (sprite) {
	case Obj_Repton:
	case Obj_Repton_Look_Left:
	case Obj_Repton_Look_Right:
	case Obj_Repton_Go_Right1:
	case Obj_Repton_Go_Right2:
	case Obj_Repton_Go_Right3:
	case Obj_Repton_Go_Right4:
	case Obj_Repton_Go_Left1:
	case Obj_Repton_Go_Left2:
	case Obj_Repton_Go_Left3:
	case Obj_Repton_Go_Left4:
	case Obj_Repton_Go_Up2:
		sprite = Obj_Repton_Go_Up1;
		break;
	case Obj_Repton_Go_Up1:
		sprite = Obj_Repton_Go_Up2;
		break;
	default:;	// Avoid warning
	}
}

// Move Repton horizontally.
void GameEngine::moveHorizontalDirect(int xd)
{
	// Move repton to the left or to the right according
	// with the xd value.
	ref(xpos, ypos) = Obj_Empty;
	ref(xpos + xd, ypos) = Obj_Repton;
	xpos += xd;

	// Change the repton sprite orientation.
	if (xd == -1)
		//  Move repton to the left.
		switch (sprite) {
		case Obj_Repton:
		case Obj_Repton_Look_Left:
		case Obj_Repton_Look_Right:
		case Obj_Repton_Go_Up1:
		case Obj_Repton_Go_Up2:
		case Obj_Repton_Go_Right1:
		case Obj_Repton_Go_Right2:
		case Obj_Repton_Go_Right3:
		case Obj_Repton_Go_Right4:
		case Obj_Repton_Go_Left4:
			sprite = Obj_Repton_Go_Left1;
			break;
		case Obj_Repton_Go_Left1:
			sprite = Obj_Repton_Go_Left2;
			break;
		case Obj_Repton_Go_Left2:
			sprite = Obj_Repton_Go_Left3;
			break;
		case Obj_Repton_Go_Left3:
			sprite = Obj_Repton_Go_Left4;
			break;
		default:;	// Avoid warning
		}
	else
		// Move repton to the right.
		switch (sprite) {
		case Obj_Repton:
		case Obj_Repton_Look_Left:
		case Obj_Repton_Look_Right:
		case Obj_Repton_Go_Up1:
		case Obj_Repton_Go_Up2:
		case Obj_Repton_Go_Left1:
		case Obj_Repton_Go_Left2:
		case Obj_Repton_Go_Left3:
		case Obj_Repton_Go_Left4:
		case Obj_Repton_Go_Right4:
			sprite = Obj_Repton_Go_Right1;
			break;
		case Obj_Repton_Go_Right1:
			sprite = Obj_Repton_Go_Right2;
			break;
		case Obj_Repton_Go_Right2:
			sprite = Obj_Repton_Go_Right3;
			break;
		case Obj_Repton_Go_Right3:
			sprite = Obj_Repton_Go_Right4;
			break;
		default:;	// Avoid warning
		}
}

// This function is called when a key is get. It turns the strongboxes into diamonds.
void GameEngine::gotKey()
{
	if (!have_key) {
		// Automagically replace all the strongboxes with diamonds.
		for (int y = 0; y < height; ++y)
			for (int x = 0; x < width; ++x)
				if (xy(x, y) == Obj_Strongbox)
					ref(x, y) = Obj_Diamond;
		have_key = true;
		if (playSound) playSound(Sound_Got_Key);
	}
}

// This function is called when Repton get an object.
void GameEngine::gotObject(SpriteObj obj)
{
	switch (obj) {
	case Obj_Diamond:
		if (playSound) playSound(Sound_Got_Diamond);
		--num_diamonds;
		score += 5;
		break;
	case Obj_TimeCapsule:
		if (playSound) playSound(Sound_Got_Time);
		game_time = capsule_time;
		break;
	case Obj_Key:
		gotKey();
		break;
	case Obj_Crown:
		if (playSound) playSound(Sound_Got_Crown);
		score += 50;
		break;
	case Obj_Bomb:
		if (num_diamonds == 0)
			status = Status_EndLevel;
		break;
	default:;	// Avoid warning
	}
}

// Move Repton to a new map location when it enters a transport.
void GameEngine::useTransporter(int x, int y)
{
	Transporter *t = findTransporter(x, y);

	// Remove the transporter and Repton from the original location.
	ref(x, y) = Obj_Empty;
	ref(xpos, ypos) = Obj_Empty;

	// Handle broken levels
	if (!t) {
		ref(x, y) = Obj_Repton;
		xpos = x;
		ypos = y;
		return;
	}

	if (showAnimation) showAnimation(Animation_Transport);
	if (playSound) playSound(Sound_Transport);

	xpos = t->dest_x;
	ypos = t->dest_y;
	switch (xy(xpos, ypos)) {
	case Obj_Diamond:
	case Obj_TimeCapsule:
	case Obj_Key:
	case Obj_Crown:
		gotObject(xy(xpos, ypos));
		// Move repton to the new location.
		ref(xpos, ypos) = Obj_Repton;
		sprite = Obj_Repton;
		break;
	case Obj_Ground1:
	case Obj_Ground2:
//		score += 2;
		/* FALLTHROUGH */
	case Obj_Empty:
		if (findMonster(xpos, ypos) != 0) {
			if (playSound) playSound(Sound_Died);
			if (showAnimation) showAnimation(Animation_Died);
			status = Status_Died;
		}
		// Move repton to the new location.
		ref(xpos, ypos) = Obj_Repton;
		sprite = Obj_Repton;
		break;
	case Obj_Transport:
		useTransporter(xpos, ypos);
		break;
	default:
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
	}
}

// Check if Repton can move vertically according with the yd value.
void GameEngine::moveVertical(int yd)
{
	SpriteObj obj = xy(xpos, ypos + yd);

	if (findMonster(xpos, ypos + yd) != 0) {
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
	}

	switch (obj) {
	case Obj_Diamond:
	case Obj_TimeCapsule:
	case Obj_Crown:
	case Obj_Key:
		gotObject(obj);
		/* FALLTHROUGH */
	case Obj_Empty:
		moveVerticalDirect(yd);
		break;
	case Obj_Ground1:
	case Obj_Ground2:
//		score += 2;
		moveVerticalDirect(yd);
		break;
	case Obj_Transport:
		useTransporter(xpos, ypos + yd);
		break;
	case Obj_Skull:
	case Obj_Fungus:
		moveVerticalDirect(yd);
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
		break;
	case Obj_Bomb:
		gotObject(Obj_Bomb);
	default:;	// Avoid warning
	}
}

// Move Repton and a near object horizontally.
void GameEngine::moveHorizontalMoveObj(int xd, SpriteObj obj)
{
	if (xy(xpos + xd*2, ypos) == Obj_Empty) {
		Monster *mp;
		if ((mp = findMonster(xpos + xd*2, ypos)) != 0 && mp->type == Obj_Monster) {
			if (playSound) playSound(Sound_Killed_Monster);
			score += 20;
			killMonster(mp);
		}
		ref(xpos + xd*2, ypos) = obj;
		moveHorizontalDirect(xd);
	}
}

// Check if Repton can move horizontally according with the xd value.
void GameEngine::moveHorizontal(int xd)
{
	SpriteObj obj = xy(xpos + xd, ypos);

	if (findMonster(xpos + xd, ypos) != 0) {
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
	}

	switch (obj) {
	case Obj_Rock:
	case Obj_Egg:
	case Obj_Broken_Egg:
	case Obj_Broken_Egg1:
	case Obj_Broken_Egg2:
	case Obj_Broken_Egg3:
		moveHorizontalMoveObj(xd, obj);
		break;
	case Obj_Diamond:
	case Obj_TimeCapsule:
	case Obj_Crown:
	case Obj_Key:
		gotObject(obj);
		/* FALLTHROUGH */
	case Obj_Empty:
		moveHorizontalDirect(xd);
		break;
	case Obj_Ground1:
	case Obj_Ground2:
//		score += 2;
		moveHorizontalDirect(xd);
		break;
	case Obj_Transport:
		useTransporter(xpos + xd, ypos);
		break;
	case Obj_Skull:
	case Obj_Fungus:
		moveHorizontalDirect(xd);
		if (playSound) playSound(Sound_Died);
		if (showAnimation) showAnimation(Animation_Died);
		status = Status_Died;
		break;
	case Obj_Bomb:
		gotObject(Obj_Bomb);
	default:;	// Avoid warning
	}
}

void GameEngine::goUp()
{
	if (!paused && status == Status_Playing)
		moveVertical(-1);
}

void GameEngine::goDown()
{
	if (!paused && status == Status_Playing)
		moveVertical(1);
}

void GameEngine::goLeft()
{
	if (!paused && status == Status_Playing)
		moveHorizontal(-1);
}

void GameEngine::goRight()
{
	if (!paused && status == Status_Playing)
		moveHorizontal(1);
}

// Kill a monster.
void GameEngine::killMonster(Monster *mp)
{
	mp->dead = true;
}

// Kill a Spirit, make it a diamond, and produce a sound.
void GameEngine::cageSpirit(Monster *mp, int x, int y)
{
	if (playSound) playSound(Sound_Cage_Spirit);
	ref(x, y) = Obj_Diamond;
	killMonster(mp);
}

void GameEngine::timer(int msecs)
{
	if (!paused && status == Status_Playing) {
		game_time -= msecs;
		if (game_time <= 0)
			status = Status_TimeOut;
		int round_time = game_time / 100;
  		if (round_time % 2 == 0)
			updateObjects();
		if (round_time % 3 == 0)
			updateMonsters();
		if (round_time % 5 == 0) {
			updateEggs();
			updateObjects();
		}
  		if (round_time % 200 == 0)
			updateFunguses();
	}
}

