/*
 * Copyright (c) 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapviewer.h"
#include "map.h"

FXDEFMAP(MapViewer) MapViewerMap[] = {
  FXMAPFUNC(SEL_MOTION,			0, MapViewer::onMotion),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,  	0, MapViewer::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,	0, MapViewer::onLeftBtnRelease),
  FXMAPFUNC(SEL_PAINT, 			0, MapViewer::onPaint),
};

FXIMPLEMENT(MapViewer, FXScrollArea, MapViewerMap, ARRAYNUMBER(MapViewerMap))

MapViewer::MapViewer()
{
}

MapViewer::MapViewer(FXComposite *p, FXObject *tgt, FXSelector sel, FXuint opts, FXint x, FXint y, FXint w, FXint h)
	: FXScrollArea(p, opts, x, y, w, h)
{
	flags |= FLAG_ENABLED;
	target = tgt;
  	message = sel;
	map = 0;
	sprites = 0;
	currenttool = Sprite_Rock;
	leftbtn = false;
	horizontalScrollbar()->setLine(sprite_width);
	horizontalScrollbar()->setPage(sprite_width);
	verticalScrollbar()->setLine(sprite_height);
	verticalScrollbar()->setPage(sprite_height);
}

MapViewer::~MapViewer()
{
}

void MapViewer::create()
{
 	FXScrollArea::create();
}

FXbool MapViewer::canFocus() const
{
	return TRUE;
}

void MapViewer::setMap(Map *m, FXBMPImage **s)
{
	map = m;
	sprites = s;
	recalc();
	update();
}

void MapViewer::setTool(Sprite tool)
{
	currenttool = tool;
}

FXint MapViewer::getContentWidth()
{
	return map ? map->width * sprite_width : 1;
}

FXint MapViewer::getContentHeight()
{
	return map ? map->height * sprite_height : 1;
}

void MapViewer::layout()
{
	// Layout scroll bars and viewport
	FXScrollArea::layout();

 	update();
	flags &= ~FLAG_DIRTY;
}

long MapViewer::onPaint(FXObject *, FXSelector, void *ptr)
{
  	FXEvent *event = (FXEvent *)ptr;
  	FXDCWindow dc(this, event);
	if (map) {
		dc.setFunction(BLT_SRC);
		int sx, sy, ex, ey;
  		sx = -pos_x / sprite_width;
  		sy = -pos_y / sprite_height;
  		ex = FXMIN(map->width, sx + viewport_w/sprite_width + 2);
  		ey = FXMIN(map->height, sy + viewport_h/sprite_height + 2);
  		for (int y = sy; y < ey; ++y)
  			for (int x = sx; x < ex; ++x)
				dc.drawImage(sprites[map->xy(x, y)], x*sprite_width+pos_x, y*sprite_height+pos_y);
		dc.setForeground(FXRGB(255, 0, 0));
   		typedef vector<Transporter *>::const_iterator I;
		for (I i = map->transporters.begin(); i != map->transporters.end(); ++i) {
			Transporter *t = *i;
  			dc.drawArc(pos_x + t->dest_x*sprite_width+2, pos_y + t->dest_y*sprite_height+2,sprite_width-6,sprite_height-6,0,360*64);
			dc.drawLine(pos_x + t->orig_x*sprite_width+sprite_width/2, pos_y + t->orig_y*sprite_height+sprite_height/2,
  				    pos_x + t->dest_x*sprite_width+sprite_width/2, pos_y + t->dest_y*sprite_height+sprite_height/2);
  		}
        	dc.setForeground(backColor);
    		dc.fillRectangle(pos_x+sprite_width*map->width, 0, viewport_w-pos_x-sprite_width*map->width, sprite_height*map->height);
    		dc.fillRectangle(0, pos_y+sprite_height*map->height, viewport_w, viewport_h-pos_y-sprite_height*map->height);
	} else {
	 	dc.setForeground(backColor);
		dc.fillRectangle(0, 0, viewport_w, viewport_h);
	}
  	return 1;
}

void MapViewer::mouseClickEvent(int x, int y)
{
	if (map->readonly) {
		leftbtn = false;
		ungrab();
		getApp()->beep();
	} else if (map->xy(x, y) != Sprite_Transporter) {
		if (currenttool == Sprite_Repton && map->count(Sprite_Repton) > 0) {
 			leftbtn = false;
    			ungrab();
			FXMessageBox::information(this, MBOX_OK, "FX-Repton Editor",
		 	    "Only one Repton can be placed.");
		} else if (currenttool == Sprite_Transporter) {
			leftbtn = false;
			ungrab();
			Transporter *t = new Transporter(x, y, x, y);
			if (getTransCoordinates(this, map, t, false)) {
				map->xy(x, y) = Sprite_Transporter;
				map->transporters.push_back(t);
				map->ismodified = true;
				if (target) target->handle(this, MKUINT(message, SEL_CHANGED), 0);
			} else
				delete t;
    		       	update();
		} else {
			map->xy(x, y) = currenttool;
    			map->ismodified = true;
		       	update();
		}
	} else {
		leftbtn = false;
		ungrab();
  		FXMessageBox::information(this, MBOX_OK, "FX-Repton Editor",
	 	    "To remove a transporter, use the Delete command in the transporters list.");
	}
}

long MapViewer::onMotion(FXObject *, FXSelector, void *ptr)
{
	FXEvent *ev = (FXEvent *)ptr;
	int x = (ev->win_x + horizontalScrollbar()->getPosition()) / sprite_width;
	int y = (ev->win_y + verticalScrollbar()->getPosition()) / sprite_height;
	if (leftbtn && x >= 0 && y >= 0 && x < map->width && y < map->height && map->xy(x, y) != currenttool) {
		if (ev->win_x < 0)
		    	setPosition(getXPosition() + sprite_width, getYPosition());
		else if (ev->win_x > viewport_w)
		    	setPosition(getXPosition() - sprite_width, getYPosition());
		if (ev->win_y < 0)
		    	setPosition(getXPosition(), getYPosition() + sprite_height);
		else if (ev->win_y > viewport_h)
		    	setPosition(getXPosition(), getYPosition() - sprite_height);
		mouseClickEvent(x, y);
	}
      	if (target && target->handle(this, MKUINT(message, SEL_MOTION), ptr)) return 1;
	return 1;
}

long MapViewer::onLeftBtnPress(FXObject *, FXSelector, void *ptr)
{
	FXEvent *ev = (FXEvent *)ptr;
	int x = (ev->win_x + horizontalScrollbar()->getPosition()) / sprite_width;
	int y = (ev->win_y + verticalScrollbar()->getPosition()) / sprite_height;
	if (x >= 0 && y >= 0 && x < map->width && y < map->height && map->xy(x, y) != currenttool) {
		leftbtn = true;
		grab();
		mouseClickEvent(x, y);
	}
	return 1;
}

long MapViewer::onLeftBtnRelease(FXObject *, FXSelector, void *)
{
	if (leftbtn) {
		leftbtn = false;
 		ungrab();
	}
	return 1;
}

bool MapViewer::getTransCoordinates(FXWindow *owner, Map *m, Transporter *t, bool change)
{
	FXDialogBox dlg(owner, (change ? "Change Transporter" : "New Transporter"), DECOR_TITLE|DECOR_BORDER,
 		0,0,0,0, 0,0,0,0, 0,0);

  	FXVerticalFrame *contents =
        	new FXVerticalFrame(&dlg, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 10,10);

 	FXGroupBox *group = new FXGroupBox(contents, "Destination Coordinates", GROUPBOX_NORMAL|FRAME_GROOVE,
  			0,0,0,0, 5,5,5,5, 10,10);

  	FXVerticalFrame *gcontents =
        	new FXVerticalFrame(group, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);

    	FXHorizontalFrame *h1 =
        	new FXHorizontalFrame(gcontents, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 10,10);

	new FXLabel(h1, "X:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
	FXSpinner *xspin = new FXSpinner(h1, 4, 0, 0, SPIN_NORMAL|FRAME_SUNKEN|FRAME_THICK);
	xspin->setRange(1, m->width);
    	xspin->setValue((change ? t->dest_x : t->orig_x)+1);

	new FXLabel(h1, "Y:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
	FXSpinner *yspin = new FXSpinner(h1, 4, 0, 0, SPIN_NORMAL|FRAME_SUNKEN|FRAME_THICK);
	yspin->setRange(1, m->height);
	yspin->setValue((change ? t->dest_y : t->orig_y)+1);

	FXHorizontalFrame *buttons =
        	new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 10,10);
        new FXButton(buttons, "Cancel", 0, &dlg,
	        FXDialogBox::ID_CANCEL, BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
              	0,0,0,0, 32,32,5,5);
	FXButton *ok = new FXButton(buttons, "OK", 0, &dlg,
	        FXDialogBox::ID_ACCEPT, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
         	0,0,0,0, 32,32,5,5);
        ok->setFocus();
        if (!dlg.execute(PLACEMENT_OWNER))
 		return false;

	t->dest_x = xspin->getValue()-1;
	t->dest_y = yspin->getValue()-1;

	return true;
}

